/*
 *	Copyright (C) 1998 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: ss.c,v 1.17 2001/09/06 06:23:21 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <sys/fcntl.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <unistd.h>

#include <geta/config.h>
#include <geta/paths.h>
#include <geta/getamu.h>
#include <geta/srvmu.h>
#include <geta/smap.h>

#include "ssP.h"

#if 1
#define DEBUG_SIGNAL
#endif

extern int errno;
extern char **environ;

static int ls = -1;
static char *localsocket = NULL;
static int type = AF_UNSPEC;

static int parentp = 1;
char *progname;

static void sigflg(int, int, int);
static void sighup(int);
static void sigchld(int);
static void sigint(int);
static void sigpipe(int);
static void die(int);
#if defined DEBUG_SIGNAL
static void sig(int);
#endif
static int init_lang_conf(char *, char *);
static int start_launcher(struct _lang_conf *, char *);
static void usage(void);
static void unlink_run(void);
static int allow(struct sockaddr_in *);

extern int nlanguages;

static char accessfile[MAXPATHLEN] = "\0";

int debug = 0;

extern struct _lang_conf *lang_conf;

struct acdata *ss_access;

static void
die(n)
{
	syslog(LOG_ERR, "ss: die(%d)", n);
	if (parentp) {
		sem_rmid();
		if (type==AF_INET) {
			close(ls);
		}
		if (type==AF_UNIX) {
			unlink(localsocket);
		}
	}
	signal(SIGINT, SIG_IGN);
	kill(0, SIGINT);
	exit(n);
}

static void
sigflg(sig, set, reset)
{
	struct sigaction sa;
	sigaction(sig, NULL, &sa);
	sa.sa_flags |= set;
	sa.sa_flags &= ~reset;
	sigaction(sig, &sa, NULL);
}

void
sighup(n)
{

	if (*accessfile) {
		syslog(LOG_INFO, "sighup: reloading accessfile");
		if (!(ss_access = init_access(accessfile))) {
			syslog(LOG_ERR, "sighup: accessfile=[%s] %s", accessfile, strerror(errno));
			die(1);
		}
	}
}

static void
sigchld(n)
{
	pid_t pid;
	int status;
	char buf[128];

	while ((pid = wait3(&status, WNOHANG, NULL)) > 0) {
		sprint_status(pid, status, 1, "SIGCHLD", buf, sizeof buf);
		if (*buf) {	/* abnormal exit */
			syslog(LOG_ERR, "sigchld: %s", buf);
		}
	}
}

static void
sigpipe(n)
{
	syslog(LOG_ERR, "sigpipe: pid=%d, is_parent=%d", getpid(), parentp);
	exit(0);
}

static void
sigint(n)
{
	syslog(LOG_ERR, "sigint: %d exit.", getpid());
	die(n);
}

#if defined SIGINFO
static void
siginfo(n)
{
	int i;
	FILE *f;
	char name[MAXPATHLEN];
	struct _lang_conf *p;

	snprintf(name, sizeof name, "/tmp/ss.%d", getpid());
	if (!(f=fopen(name, "a"))) {
		syslog(LOG_ERR, "siginfo: %s %s", name, strerror(errno));
		return;
	}
	for (i=0, p=lang_conf; i<nlanguages; p++, i++) {
		fprintf(f, "--- name = %s\n", p->language);
		fprintf(f, "--- semidx = %d\n", p->semidx);
		sem_status(p->semidx, f);
	}
	fclose(f);
	syslog(LOG_ERR, "siginfo: profile saved in %s", name);
}
#endif

#if defined DEBUG_SIGNAL
static void
sig(n)
{
	syslog(LOG_ERR, "caught a signal %d", n);
	exit(1);
}
#endif

#define	BACKLOG	16

static void
usage()
{
	fprintf(stderr, "Usage: %s [-d|-daemon] [-D debug-level] (-l local_socket|-i port)\n", progname);
	die(1);
}

static char run[MAXPATHLEN];

static void
unlink_run()
{
	if (parentp) {
		unlink(run);
	}
}

main(argc, argv)
char *argv[];
{
	struct stat sb;
	char *projectroot;
	int port=0;
#if defined DEBUG_SIGNAL
	int i;
#endif
	struct sockaddr *peeraddr = NULL;
#if HAVE_SOCKLEN_T
	socklen_t addrlen;
#else
	size_t addrlen;
#endif
	struct sockaddr_un peeraddr_un;
	struct sockaddr_in peeraddr_in;
	FILE *f;
	int detache = 0;
	static char configfile[MAXPATHLEN];

	if (setpgid(0, getpid())==-1) {
		perror("setpgid");
		exit(1);
	}
	if (!(projectroot = getenv(PROOTENV))) {
		fprintf(stderr, "warning: environment variable " PROOTENV " is not set\n");
	}

	if (progname=rindex(argv[0], '/')) {
		progname++;
	}
	else {
		progname = argv[0];
	}

#if defined DEBUG_SIGNAL
	for (i=0; i<32; i++) {
		signal(i, sig);
	}
#endif
	openlog(progname, LOG_PID, LOG_LOCAL1);
#if 0
	setlogmask(LOG_UPTO(LOG_INFO));
#endif
	syslog(LOG_INFO, "smap started %d", getpid());

	snprintf(run, sizeof run, "%s" SMAP_FN_RUNFILE, projectroot);
	if (stat(run, &sb)!=-1) {
		fprintf(stderr, "%s exists.\n", run);
		exit(1);
	}

#if 0
	climit();
#endif

	type = AF_UNSPEC;

	for (argc--, argv++; argc>0; argc--, argv++) {
		if (*argv[0]!='-') break;
		if (!strcmp(argv[0], "-d") || !strcmp(argv[0], "-daemon")) {
			detache = 1;
		}
		else if (!strcmp(argv[0], "-D")) {
			if (argc<=1) {
				usage();
			}
			argc--, argv++;
			debug = strtoul(argv[0], NULL, 10);
			fprintf(stderr, "Debug level = %d\n", debug);
		}
		else if (!strcmp(argv[0], "-l")) {
			if (type!=AF_UNSPEC || argc<=1) {
				usage();
			}
			argc--, argv++;
			localsocket = argv[0];
			fprintf(stderr, "local socket = %s\n", localsocket);
			type = AF_UNIX;
		}
		else if (!strcmp(argv[0], "-i")) {
			if (type!=AF_UNSPEC || argc<=1) {
				usage();
			}
			argc--, argv++;
			if ((port = atoi(*argv))==0) {
				fprintf(stderr, "ss: error: port==0\n");
				die(1);
			}
			fprintf(stderr, "port = %d\n", port);
			type = AF_INET;
		}
		else {
			usage();
		}
	}

	if (argc!=0 || type==AF_UNSPEC) {
		usage();
	}
	if (type==AF_INET) {
		snprintf(accessfile, sizeof accessfile, "%s" SMAP_FN_ACCESSFILE, projectroot);
	}

	snprintf(configfile, sizeof configfile, "%s" SMAP_FN_CONFIGFILE, projectroot);

	if (type==AF_INET && !(ss_access = init_access(accessfile))) {
		perror(accessfile);
		die(1);
	}

	if (init_lang_conf(configfile, projectroot)==-1) {
		die(1);
	}

	if (type==AF_UNIX) {
		if ((ls=un_sbl(localsocket, BACKLOG))==-1) {
			return -1;
		}
		chmod(localsocket, 0660);
	}
	else if (type==AF_INET) {
		if ((ls=in_sbl(port, BACKLOG))==-1) {
			return -1;
		}
	}
	else {
		fprintf(stderr, "ss: internal error\n");
		die(1);
	}

	if (sem_create(nlanguages)==-1) {
		syslog(LOG_ERR, "%s: sem_create: cannot create a semaphore", progname);
		sem_status(0, stderr);
		die(1);
	}

	if (detache) {
		detachetty();
	}

	if (f=fopen(run, "w")) {
		fprintf(f, "%d", (int)getpid());
		fclose(f);
		atexit(unlink_run);
	}

	signal(SIGINT, sigint);
	signal(SIGCHLD, sigchld);
	sigflg(SIGCHLD, SA_RESTART, SA_RESETHAND);
	signal(SIGTERM, SIG_IGN);
	signal(SIGPIPE, sigpipe);
	sigflg(SIGPIPE, 0, SA_RESETHAND);
#if defined SIGINFO
	signal(SIGINFO, siginfo);
	sigflg(SIGINFO, SA_RESTART, SA_RESETHAND);
#endif
	signal(SIGHUP, sighup);
	sigflg(SIGHUP, SA_RESTART, SA_RESETHAND);

	for (;;) {
		int s = -1;
		if (type==AF_UNIX) {
			addrlen = sizeof peeraddr_un;
			bzero(&peeraddr_un, sizeof peeraddr_un);
			peeraddr = (struct sockaddr *)&peeraddr_un;
		}
		else if (type==AF_INET) {
			addrlen = sizeof peeraddr_in;
			bzero(&peeraddr_in, sizeof peeraddr_in);
			peeraddr = (struct sockaddr *)&peeraddr_in;
		}
		else {
			fprintf(stderr, "ss: internal error\n");
			die(1);
		}
		if ((s = accept(ls, peeraddr, &addrlen))==-1) {
			if (errno!=EINTR) {
				perror("accept");
				sleep(1);
			}
else perror("SS: accept");
			continue;
		}
		if (type==AF_INET) {
			if (!allow(&peeraddr_in)) {
				syslog(LOG_ERR, "%s: access denied", inet_ntoa(peeraddr_in.sin_addr));
				continue;
			}
		}
#if defined DEBUG_FORK
		switch (parentp = 0) {
#else
		switch (parentp = fork()) {
#endif
		case -1:
			syslog(LOG_ERR, "fork %s", strerror(errno));
			die(1);
			break;
		case 0: {
#if defined SIGINFO
			signal(SIGINFO, SIG_IGN);
#endif
			signal(SIGHUP, SIG_IGN);

			close(ls);
			do_server(s);
			exit(0);
		}
		default:
			close(s);
			break;
		}
	}
}

static int
init_lang_conf(configfile, projectroot)
char *configfile, *projectroot;
{
	int i;
	struct _lang_conf *c;

	if (readconf(configfile, projectroot)==-1) {
		return -1;
	}

	for (i=0; i<nlanguages; i++) {
		c = &lang_conf[i];
		c->fds[0] = c->fds[1] = -1;
		c->semidx = i;
		c->pid = -1;
		if (start_launcher(c, projectroot)==-1) {
			syslog(LOG_ERR, "failed to launch %s", c->language);
			return -1;
		}
	}
	return 0;
}

static
start_launcher(c, projectroot)
struct _lang_conf *c;
char *projectroot;
{
/*
   main.c -> fds[1]==fds0[1]--pipe0--fds0[0]==0 -> external program
   main.c <- fds[0]==fds1[0]--pipe1--fds1[1]==1 <- external program
*/
	int fds0[2], fds1[2];

	if (pipe(fds0)==-1) return -1;
	if (pipe(fds1)==-1) return -1;
	switch (c->pid = fork()) {
		char *av[2];
	case -1:
		syslog(LOG_ERR, "fork %s", strerror(errno));
		return -1;
		break;
	case 0:
#if defined SIGINFO
		signal(SIGINFO, SIG_IGN);
#endif
		signal(SIGHUP, SIG_IGN);

		close(fds0[1]);
		close(fds1[0]);
		if (dup2(fds0[0], 0)==-1) {
			perror("dup2");
			return -1;
		}
		close(fds0[0]);
		if (dup2(fds1[1], 1)==-1) {
			perror("dup2");
			return -1;
		}
		close(fds1[1]);

		if (c->start) {
#if defined EXEC_ON_DEMAND
			fd_set readfds;
#endif
			av[0] = c->start;
			av[1] = NULL;

#if defined EXEC_ON_DEMAND
			FD_ZERO(&readfds);
			FD_SET(0, &readfds);
			while (select(1, &readfds, NULL, NULL, NULL)==-1) ;
#endif

			execve(c->start, av, environ);
			perror("execve");
		}
		else if (c->analyse) {
			(*c->analyse)();
			syslog(LOG_ERR, "error: \"%s\" returns", c->language);
		}
		else {
			fprintf(stderr, "ss: configuration error lang=\"%s\"\n", c->language);
		}

		exit(1);
	default:
		close(fds0[0]);
		close(fds1[1]);
		c->fds[0] = fds1[0];
		c->fds[1] = fds0[1];
		return 0;
	}
}

void
sstate(name)
char *name;
{
	if (debug & D_SSTATE) {
		syslog(LOG_INFO, "sstate: %s", name);
	}
}

static int
allow(struct sockaddr_in *peer)
{
	return access_check(peer->sin_addr.s_addr, ss_access);
}
